# Copyright 2005-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from _emerge.Package import Package

import portage
from portage import os
from portage.exception import InvalidData
from portage.versions import _pkg_str


class MoveHandler:
    def __init__(self, tree, porttree):
        self._tree = tree
        self._portdb = porttree.dbapi
        self._update_keys = Package._dep_keys
        self._master_repo = self._portdb.repositories.mainRepo()
        if self._master_repo is not None:
            self._master_repo = self._master_repo.name

    def _grab_global_updates(self):
        from portage.update import grab_updates, parse_updates

        retupdates = {}
        errors = []

        for repo_name in self._portdb.getRepositories():
            repo = self._portdb.getRepositoryPath(repo_name)
            updpath = os.path.join(repo, "profiles", "updates")
            if not os.path.isdir(updpath):
                continue

            try:
                rawupdates = grab_updates(updpath)
            except portage.exception.DirectoryNotFound:
                rawupdates = []
            upd_commands = []
            for mykey, mystat, mycontent in rawupdates:
                commands, errors = parse_updates(mycontent)
                upd_commands.extend(commands)
                errors.extend(errors)
            retupdates[repo_name] = upd_commands

        if self._master_repo in retupdates:
            retupdates["DEFAULT"] = retupdates[self._master_repo]

        return retupdates, errors

    def check(self, **kwargs):
        onProgress = kwargs.get("onProgress", None)
        allupdates, errors = self._grab_global_updates()
        # Matching packages and moving them is relatively fast, so the
        # progress bar is updated in indeterminate mode.
        match = self._tree.dbapi.match
        aux_get = self._tree.dbapi.aux_get
        pkg_str = self._tree.dbapi._pkg_str
        settings = self._tree.dbapi.settings
        if onProgress:
            onProgress(0, 0)
        for repo, updates in allupdates.items():
            if repo == "DEFAULT":
                continue
            if not updates:
                continue

            def repo_match(repository):
                return repository == repo or (
                    repo == self._master_repo and repository not in allupdates
                )

            for i, update_cmd in enumerate(updates):
                if update_cmd[0] == "move":
                    origcp, newcp = update_cmd[1:]
                    for cpv in match(origcp):
                        try:
                            cpv = pkg_str(cpv, origcp.repo)
                        except (KeyError, InvalidData):
                            continue
                        if repo_match(cpv.repo):
                            build_time = getattr(cpv, "build_time", None)
                            if build_time is not None:
                                # If this update has already been applied to the same
                                # package build then silently continue.
                                for maybe_applied in match(
                                    "={}".format(cpv.replace(cpv.cp, str(newcp), 1))
                                ):
                                    if maybe_applied.build_time == build_time:
                                        break
                                else:
                                    errors.append("'%s' moved to '%s'" % (cpv, newcp))
                elif update_cmd[0] == "slotmove":
                    pkg, origslot, newslot = update_cmd[1:]
                    atom = pkg.with_slot(origslot)
                    for cpv in match(atom):
                        try:
                            cpv = pkg_str(cpv, atom.repo)
                        except (KeyError, InvalidData):
                            continue
                        if repo_match(cpv.repo):
                            errors.append(
                                "'%s' slot moved from '%s' to '%s'"
                                % (cpv, origslot, newslot)
                            )
                if onProgress:
                    onProgress(0, 0)

        # Searching for updates in all the metadata is relatively slow, so this
        # is where the progress bar comes out of indeterminate mode.
        cpv_all = self._tree.dbapi.cpv_all()
        cpv_all.sort()
        maxval = len(cpv_all)
        meta_keys = self._update_keys + self._portdb._pkg_str_aux_keys
        if onProgress:
            onProgress(maxval, 0)
        for i, cpv in enumerate(cpv_all):
            try:
                metadata = dict(zip(meta_keys, aux_get(cpv, meta_keys)))
            except KeyError:
                continue
            try:
                pkg = _pkg_str(cpv, metadata=metadata, settings=settings)
            except InvalidData:
                continue
            metadata = dict((k, metadata[k]) for k in self._update_keys)
            try:
                updates = allupdates[pkg.repo]
            except KeyError:
                try:
                    updates = allupdates["DEFAULT"]
                except KeyError:
                    continue
            if not updates:
                continue
            metadata_updates = portage.update_dbentries(updates, metadata, parent=pkg)
            if metadata_updates:
                errors.append("'%s' has outdated metadata" % cpv)
            if onProgress:
                onProgress(maxval, i + 1)

        if errors:
            return (False, errors)
        return (True, None)

    def fix(self, **kwargs):
        onProgress = kwargs.get("onProgress", None)
        allupdates, errors = self._grab_global_updates()
        # Matching packages and moving them is relatively fast, so the
        # progress bar is updated in indeterminate mode.
        move = self._tree.dbapi.move_ent
        slotmove = self._tree.dbapi.move_slot_ent
        if onProgress:
            onProgress(0, 0)
        for repo, updates in allupdates.items():
            if repo == "DEFAULT":
                continue
            if not updates:
                continue

            def repo_match(repository):
                return repository == repo or (
                    repo == self._master_repo and repository not in allupdates
                )

            for i, update_cmd in enumerate(updates):
                if update_cmd[0] == "move":
                    move(update_cmd, repo_match=repo_match)
                elif update_cmd[0] == "slotmove":
                    slotmove(update_cmd, repo_match=repo_match)
                if onProgress:
                    onProgress(0, 0)

        # Searching for updates in all the metadata is relatively slow, so this
        # is where the progress bar comes out of indeterminate mode.
        self._tree.dbapi.update_ents(allupdates, onProgress=onProgress)
        if errors:
            return (False, errors)
        return (True, None)


class MoveInstalled(MoveHandler):

    short_desc = "Perform package move updates for installed packages"

    @staticmethod
    def name():
        return "moveinst"

    def __init__(self):
        eroot = portage.settings["EROOT"]
        MoveHandler.__init__(
            self, portage.db[eroot]["vartree"], portage.db[eroot]["porttree"]
        )


class MoveBinary(MoveHandler):

    short_desc = "Perform package move updates for binary packages"

    @staticmethod
    def name():
        return "movebin"

    def __init__(self):
        eroot = portage.settings["EROOT"]
        MoveHandler.__init__(
            self, portage.db[eroot]["bintree"], portage.db[eroot]["porttree"]
        )



# Copyright 1999-2021 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

from collections import deque

from portage.util.futures import asyncio
from portage.util.SlotObject import SlotObject


class SequentialTaskQueue(SlotObject):

    __slots__ = ("max_jobs", "running_tasks") + ("_scheduling", "_task_queue")

    def __init__(self, **kwargs):
        SlotObject.__init__(self, **kwargs)
        self._task_queue = deque()
        self.running_tasks = set()
        if self.max_jobs is None:
            self.max_jobs = 1

    def add(self, task):
        self._task_queue.append(task)
        self.schedule()

    def addFront(self, task):
        self._task_queue.appendleft(task)
        self.schedule()

    def schedule(self):

        if self._scheduling:
            # Ignore any recursive schedule() calls triggered via
            # self._task_exit().
            return

        self._scheduling = True
        try:
            while self._task_queue and (
                self.max_jobs is True or len(self.running_tasks) < self.max_jobs
            ):
                task = self._task_queue.popleft()
                cancelled = getattr(task, "cancelled", None)
                if not cancelled:
                    self.running_tasks.add(task)
                    task.addExitListener(self._task_exit)
                    task.start()
        finally:
            self._scheduling = False

    def _task_exit(self, task):
        """
        Since we can always rely on exit listeners being called, the set of
        running tasks is always pruned automatically and there is never any need
        to actively prune it.
        """
        self.running_tasks.remove(task)
        if self._task_queue:
            self.schedule()

    def clear(self):
        """
        Clear the task queue and asynchronously terminate any running tasks.
        """
        for task in self._task_queue:
            task.cancel()
        self._task_queue.clear()

        for task in list(self.running_tasks):
            task.cancel()

    async def wait(self, loop=None):
        """
        Wait for the queue to become empty. This method is a coroutine.
        """
        while self:
            task = next(iter(self.running_tasks), None)
            if task is None:
                # Wait for self.running_tasks to populate.
                await asyncio.sleep(0, loop=loop)
            else:
                await task.async_wait()

    def __bool__(self):
        return bool(self._task_queue or self.running_tasks)

    def __len__(self):
        return len(self._task_queue) + len(self.running_tasks)



# Copyright: 2005 Gentoo Foundation
# Author(s): Brian Harring (ferringb@gentoo.org)
# License: GPL2


class CacheError(Exception):
    pass


class InitializationError(CacheError):
    def __init__(self, class_name, error):
        self.error, self.class_name = error, class_name

    def __str__(self):
        return "Creation of instance %s failed due to %s" % (
            self.class_name,
            str(self.error),
        )


class CacheCorruption(CacheError):
    def __init__(self, key, ex):
        self.key, self.ex = key, ex

    def __str__(self):
        return "%s is corrupt: %s" % (self.key, str(self.ex))


class GeneralCacheCorruption(CacheError):
    def __init__(self, ex):
        self.ex = ex

    def __str__(self):
        return "corruption detected: %s" % str(self.ex)


class InvalidRestriction(CacheError):
    def __init__(self, key, restriction, exception=None):
        if exception is None:
            exception = ""
        self.key, self.restriction, self.ex = key, restriction, ex

    def __str__(self):
        return "%s:%s is not valid: %s" % (self.key, self.restriction, str(self.ex))


class ReadOnlyRestriction(CacheError):
    def __init__(self, info=""):
        self.info = info

    def __str__(self):
        return "cache is non-modifiable" + str(self.info)


class StatCollision(CacheError):
    """
    If the content of a cache entry changes and neither the file mtime nor
    size changes, it will prevent rsync from detecting changes. Cache backends
    may raise this exception from _setitem() if they detect this type of stat
    collision. See bug #139134.
    """

    def __init__(self, key, filename, mtime, size):
        self.key = key
        self.filename = filename
        self.mtime = mtime
        self.size = size

    def __str__(self):
        return "%s has stat collision with size %s and mtime %s" % (
            self.key,
            self.size,
            self.mtime,
        )

    def __repr__(self):
        return "portage.cache.cache_errors.StatCollision(%s)" % (
            ", ".join(
                (repr(self.key), repr(self.filename), repr(self.mtime), repr(self.size))
            ),
        )



# elog/messages.py - elog core functions
# Copyright 2006-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2

import portage

portage.proxy.lazyimport.lazyimport(
    globals(),
    "portage.output:colorize",
    "portage.util:writemsg",
)

from portage.const import EBUILD_PHASES
from portage.localization import _
from portage import os
from portage import _encodings
from portage import _unicode_encode
from portage import _unicode_decode

import io
import sys

_log_levels = frozenset(
    [
        "ERROR",
        "INFO",
        "LOG",
        "QA",
        "WARN",
    ]
)


def collect_ebuild_messages(path):
    """Collect elog messages generated by the bash logging function stored
    at 'path'.
    """
    mylogfiles = None
    try:
        mylogfiles = os.listdir(path)
    except OSError:
        pass
    # shortcut for packages without any messages
    if not mylogfiles:
        return {}
    # exploit listdir() file order so we process log entries in chronological order
    mylogfiles.reverse()
    logentries = {}
    for msgfunction in mylogfiles:
        filename = os.path.join(path, msgfunction)
        if msgfunction not in EBUILD_PHASES:
            writemsg(
                _("!!! can't process invalid log file: %s\n") % filename, noiselevel=-1
            )
            continue
        if not msgfunction in logentries:
            logentries[msgfunction] = []
        lastmsgtype = None
        msgcontent = []
        f = io.open(
            _unicode_encode(filename, encoding=_encodings["fs"], errors="strict"),
            mode="r",
            encoding=_encodings["repo.content"],
            errors="replace",
        )
        # Use split('\n') since normal line iteration or readlines() will
        # split on \r characters as shown in bug #390833.
        for l in f.read().split("\n"):
            if not l:
                continue
            try:
                msgtype, msg = l.split(" ", 1)
                if msgtype not in _log_levels:
                    raise ValueError(msgtype)
            except ValueError:
                writemsg(
                    _("!!! malformed entry in " "log file: '%s': %s\n") % (filename, l),
                    noiselevel=-1,
                )
                continue

            if lastmsgtype is None:
                lastmsgtype = msgtype

            if msgtype == lastmsgtype:
                msgcontent.append(msg)
            else:
                if msgcontent:
                    logentries[msgfunction].append((lastmsgtype, msgcontent))
                msgcontent = [msg]
            lastmsgtype = msgtype
        f.close()
        if msgcontent:
            logentries[msgfunction].append((lastmsgtype, msgcontent))

    # clean logfiles to avoid repetitions
    for f in mylogfiles:
        try:
            os.unlink(os.path.join(path, f))
        except OSError:
            pass
    return logentries


_msgbuffer = {}


def _elog_base(level, msg, phase="other", key=None, color=None, out=None):
    """Backend for the other messaging functions, should not be called
    directly.
    """

    # TODO: Have callers pass in a more unique 'key' parameter than a plain
    # cpv, in order to ensure that messages are properly grouped together
    # for a given package instance, and also to ensure that each elog module's
    # process() function is only called once for each unique package. This is
    # needed not only when building packages in parallel, but also to preserve
    # continuity in messages when a package is simply updated, since we don't
    # want the elog_process() call from the uninstall of the old version to
    # cause discontinuity in the elog messages of the new one being installed.

    global _msgbuffer

    if out is None:
        out = sys.stdout

    if color is None:
        color = "INFO"

    msg = _unicode_decode(msg, encoding=_encodings["content"], errors="replace")

    formatted_msg = colorize(color, " * ") + msg + "\n"

    # avoid potential UnicodeEncodeError
    if out in (sys.stdout, sys.stderr):
        formatted_msg = _unicode_encode(
            formatted_msg, encoding=_encodings["stdio"], errors="backslashreplace"
        )
        out = out.buffer

    out.write(formatted_msg)

    if key not in _msgbuffer:
        _msgbuffer[key] = {}
    if phase not in _msgbuffer[key]:
        _msgbuffer[key][phase] = []
    _msgbuffer[key][phase].append((level, msg))

    # raise NotImplementedError()


def collect_messages(key=None, phasefilter=None):
    global _msgbuffer

    if key is None:
        rValue = _msgbuffer
        _reset_buffer()
    else:
        rValue = {}
        if key in _msgbuffer:
            if phasefilter is None:
                rValue[key] = _msgbuffer.pop(key)
            else:
                rValue[key] = {}
                for phase in phasefilter:
                    try:
                        rValue[key][phase] = _msgbuffer[key].pop(phase)
                    except KeyError:
                        pass
                if not _msgbuffer[key]:
                    del _msgbuffer[key]
    return rValue


def _reset_buffer():
    """Reset the internal message buffer when it has been processed,
    should not be called directly.
    """
    global _msgbuffer

    _msgbuffer = {}


# creating and exporting the actual messaging functions
_functions = {
    "einfo": ("INFO", "INFO"),
    "elog": ("LOG", "LOG"),
    "ewarn": ("WARN", "WARN"),
    "eqawarn": ("QA", "QAWARN"),
    "eerror": ("ERROR", "ERR"),
}


class _make_msgfunction:
    __slots__ = ("_color", "_level")

    def __init__(self, level, color):
        self._level = level
        self._color = color

    def __call__(self, msg, phase="other", key=None, out=None):
        """
        Display and log a message assigned to the given key/cpv.
        """
        _elog_base(self._level, msg, phase=phase, key=key, color=self._color, out=out)


for f in _functions:
    setattr(
        sys.modules[__name__], f, _make_msgfunction(_functions[f][0], _functions[f][1])
    )
del f, _functions



# Copyright 2015 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

from __future__ import print_function
import sys

from portage.const import SUPPORTED_GENTOO_BINPKG_FORMATS
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import (
    ResolverPlayground,
    ResolverPlaygroundTestCase,
)
from portage.output import colorize


class SonameAutoUnmaskTestCase(TestCase):
    def testSonameAutoUnmask(self):

        binpkgs = {
            "dev-libs/icu-49": {
                "KEYWORDS": "x86",
                "PROVIDES": "x86_32: libicu.so.49",
            },
            "dev-libs/icu-4.8": {
                "KEYWORDS": "x86",
                "PROVIDES": "x86_32: libicu.so.48",
            },
            "dev-libs/libxml2-2.7.8": {
                "KEYWORDS": "~x86",
                "DEPEND": "dev-libs/icu",
                "RDEPEND": "dev-libs/icu",
                "REQUIRES": "x86_32: libicu.so.49",
            },
        }

        installed = {
            "dev-libs/icu-4.8": {
                "KEYWORDS": "x86",
                "PROVIDES": "x86_32: libicu.so.48",
            },
            "dev-libs/libxml2-2.7.8": {
                "KEYWORDS": "~x86",
                "DEPEND": "dev-libs/icu",
                "RDEPEND": "dev-libs/icu",
                "REQUIRES": "x86_32: libicu.so.48",
            },
        }

        world = ["dev-libs/libxml2"]

        test_cases = (
            ResolverPlaygroundTestCase(
                ["dev-libs/icu"],
                options={
                    "--autounmask": True,
                    "--ignore-soname-deps": "n",
                    "--oneshot": True,
                    "--usepkgonly": True,
                },
                success=False,
                mergelist=["[binary]dev-libs/icu-49", "[binary]dev-libs/libxml2-2.7.8"],
                unstable_keywords=["dev-libs/libxml2-2.7.8"],
            ),
            ResolverPlaygroundTestCase(
                ["dev-libs/icu"],
                options={
                    "--autounmask": True,
                    "--ignore-soname-deps": "y",
                    "--oneshot": True,
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=["[binary]dev-libs/icu-49"],
            ),
            # Test that dev-libs/icu-49 update is skipped due to
            # dev-libs/libxml2-2.7.8 being masked by KEYWORDS. Note
            # that this result is questionable, since the installed
            # dev-libs/libxml2-2.7.8 instance is also masked!
            ResolverPlaygroundTestCase(
                ["@world"],
                options={
                    "--autounmask": True,
                    "--deep": True,
                    "--ignore-soname-deps": "n",
                    "--update": True,
                    "--usepkgonly": True,
                },
                success=True,
                mergelist=[],
            ),
        )

        for binpkg_format in SUPPORTED_GENTOO_BINPKG_FORMATS:
            with self.subTest(binpkg_format=binpkg_format):
                print(colorize("HILITE", binpkg_format), end=" ... ")
                sys.stdout.flush()
                playground = ResolverPlayground(
                    binpkgs=binpkgs,
                    installed=installed,
                    world=world,
                    debug=False,
                    user_config={
                        "make.conf": ('BINPKG_FORMAT="%s"' % binpkg_format,),
                    },
                )

                try:
                    for test_case in test_cases:
                        playground.run_TestCase(test_case)
                        self.assertEqual(
                            test_case.test_success, True, test_case.fail_msg
                        )
                finally:
                    playground.debug = False
                    playground.cleanup()



# Copyright 1999-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import portage
from portage import os
from portage.elog.messages import eerror
from portage.util.SlotObject import SlotObject


class EbuildFetchonly(SlotObject):

    __slots__ = ("fetch_all", "pkg", "pretend", "settings")

    def execute(self):
        settings = self.settings
        pkg = self.pkg
        portdb = pkg.root_config.trees["porttree"].dbapi
        ebuild_path = portdb.findname(pkg.cpv, myrepo=pkg.repo)
        if ebuild_path is None:
            raise AssertionError("ebuild not found for '%s'" % pkg.cpv)
        settings.setcpv(pkg)
        debug = settings.get("PORTAGE_DEBUG") == "1"

        rval = portage.doebuild(
            ebuild_path,
            "fetch",
            settings=settings,
            debug=debug,
            listonly=self.pretend,
            fetchonly=1,
            fetchall=self.fetch_all,
            mydbapi=portdb,
            tree="porttree",
        )

        # For pretend mode, this error message is suppressed,
        # and the unsuccessful return value is used to trigger
        # a call to the pkg_nofetch phase.
        if rval != os.EX_OK and not self.pretend:
            msg = "Fetch failed for '%s'" % (pkg.cpv,)
            eerror(msg, phase="unpack", key=pkg.cpv)

        return rval



# Copyright 2012-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import subprocess
import sys
import textwrap

import portage
from portage import os
from portage import _unicode_decode
from portage.const import BASH_BINARY, PORTAGE_PYM_PATH, USER_CONFIG_PATH
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util import ensure_dirs


class PortdbCacheTestCase(TestCase):
    def testPortdbCache(self):
        debug = False

        ebuilds = {
            "dev-libs/A-1": {},
            "dev-libs/A-2": {},
            "sys-apps/B-1": {},
            "sys-apps/B-2": {},
        }

        playground = ResolverPlayground(ebuilds=ebuilds, debug=debug)
        settings = playground.settings
        eprefix = settings["EPREFIX"]
        test_repo_location = settings.repositories["test_repo"].location
        user_config_dir = os.path.join(eprefix, USER_CONFIG_PATH)
        metadata_dir = os.path.join(test_repo_location, "metadata")
        md5_cache_dir = os.path.join(metadata_dir, "md5-cache")
        pms_cache_dir = os.path.join(metadata_dir, "cache")
        layout_conf_path = os.path.join(metadata_dir, "layout.conf")

        portage_python = portage._python_interpreter
        egencache_cmd = (
            portage_python,
            "-b",
            "-Wd",
            os.path.join(self.bindir, "egencache"),
            "--update-manifests",
            "--sign-manifests=n",
            "--repo",
            "test_repo",
            "--repositories-configuration",
            settings.repositories.config_string(),
        )
        python_cmd = (portage_python, "-b", "-Wd", "-c")

        test_commands = (
            (lambda: not os.path.exists(pms_cache_dir),),
            (lambda: not os.path.exists(md5_cache_dir),),
            python_cmd
            + (
                textwrap.dedent(
                    """
				import os, sys, portage
				if portage.portdb.repositories['test_repo'].location in portage.portdb._pregen_auxdb:
					sys.exit(1)
			"""
                ),
            ),
            egencache_cmd + ("--update",),
            (lambda: not os.path.exists(pms_cache_dir),),
            (lambda: os.path.exists(md5_cache_dir),),
            python_cmd
            + (
                textwrap.dedent(
                    """
				import os, sys, portage
				if portage.portdb.repositories['test_repo'].location not in portage.portdb._pregen_auxdb:
					sys.exit(1)
			"""
                ),
            ),
            python_cmd
            + (
                textwrap.dedent(
                    """
				import os, sys, portage
				from portage.cache.flat_hash import md5_database
				if not isinstance(portage.portdb._pregen_auxdb[portage.portdb.repositories['test_repo'].location], md5_database):
					sys.exit(1)
			"""
                ),
            ),
            (
                BASH_BINARY,
                "-c",
                "echo %s > %s"
                % tuple(
                    map(
                        portage._shell_quote,
                        (
                            "cache-formats = md5-dict pms",
                            layout_conf_path,
                        ),
                    )
                ),
            ),
            egencache_cmd + ("--update",),
            (lambda: os.path.exists(md5_cache_dir),),
            python_cmd
            + (
                textwrap.dedent(
                    """
				import os, sys, portage
				if portage.portdb.repositories['test_repo'].location not in portage.portdb._pregen_auxdb:
					sys.exit(1)
			"""
                ),
            ),
            python_cmd
            + (
                textwrap.dedent(
                    """
				import os, sys, portage
				from portage.cache.flat_hash import md5_database
				if not isinstance(portage.portdb._pregen_auxdb[portage.portdb.repositories['test_repo'].location], md5_database):
					sys.exit(1)
			"""
                ),
            ),
            # Disable DeprecationWarnings, since the pms format triggers them
            # in portdbapi._create_pregen_cache().
            (
                BASH_BINARY,
                "-c",
                "echo %s > %s"
                % tuple(
                    map(
                        portage._shell_quote,
                        (
                            "cache-formats = pms md5-dict",
                            layout_conf_path,
                        ),
                    )
                ),
            ),
            (portage_python, "-b", "-Wd", "-Wi::DeprecationWarning", "-c")
            + (
                textwrap.dedent(
                    """
				import os, sys, portage
				if portage.portdb.repositories['test_repo'].location not in portage.portdb._pregen_auxdb:
					sys.exit(1)
			"""
                ),
            ),
            (portage_python, "-b", "-Wd", "-Wi::DeprecationWarning", "-c")
            + (
                textwrap.dedent(
                    """
				import os, sys, portage
				from portage.cache.metadata import database as pms_database
				if not isinstance(portage.portdb._pregen_auxdb[portage.portdb.repositories['test_repo'].location], pms_database):
					sys.exit(1)
			"""
                ),
            ),
            # Test auto-detection and preference for md5-cache when both
            # cache formats are available but layout.conf is absent.
            (BASH_BINARY, "-c", "rm %s" % portage._shell_quote(layout_conf_path)),
            python_cmd
            + (
                textwrap.dedent(
                    """
				import os, sys, portage
				if portage.portdb.repositories['test_repo'].location not in portage.portdb._pregen_auxdb:
					sys.exit(1)
			"""
                ),
            ),
            python_cmd
            + (
                textwrap.dedent(
                    """
				import os, sys, portage
				from portage.cache.flat_hash import md5_database
				if not isinstance(portage.portdb._pregen_auxdb[portage.portdb.repositories['test_repo'].location], md5_database):
					sys.exit(1)
			"""
                ),
            ),
        )

        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is not None and not pythonpath.strip():
            pythonpath = None
        if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
            pass
        else:
            if pythonpath is None:
                pythonpath = ""
            else:
                pythonpath = ":" + pythonpath
            pythonpath = PORTAGE_PYM_PATH + pythonpath

        env = {
            "PATH": os.environ.get("PATH", ""),
            "PORTAGE_OVERRIDE_EPREFIX": eprefix,
            "PORTAGE_PYTHON": portage_python,
            "PORTAGE_REPOSITORIES": settings.repositories.config_string(),
            "PYTHONDONTWRITEBYTECODE": os.environ.get("PYTHONDONTWRITEBYTECODE", ""),
            "PYTHONPATH": pythonpath,
        }

        if "__PORTAGE_TEST_HARDLINK_LOCKS" in os.environ:
            env["__PORTAGE_TEST_HARDLINK_LOCKS"] = os.environ[
                "__PORTAGE_TEST_HARDLINK_LOCKS"
            ]

        dirs = [user_config_dir]

        try:
            for d in dirs:
                ensure_dirs(d)

            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for i, args in enumerate(test_commands):

                if hasattr(args[0], "__call__"):
                    self.assertTrue(args[0](), "callable at index %s failed" % (i,))
                    continue

                proc = subprocess.Popen(args, env=env, stdout=stdout)

                if debug:
                    proc.wait()
                else:
                    output = proc.stdout.readlines()
                    proc.wait()
                    proc.stdout.close()
                    if proc.returncode != os.EX_OK:
                        for line in output:
                            sys.stderr.write(_unicode_decode(line))

                self.assertEqual(
                    os.EX_OK,
                    proc.returncode,
                    "command %d failed with args %s"
                    % (
                        i,
                        args,
                    ),
                )
        finally:
            playground.cleanup()



# Copyright 2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import io

from _emerge.CompositeTask import CompositeTask
from _emerge.EbuildProcess import EbuildProcess
from _emerge.SpawnProcess import SpawnProcess

import portage
from portage import os
from portage import _encodings
from portage import _unicode_encode
from portage.util._async.AsyncFunction import AsyncFunction
from portage.util.install_mask import install_mask_dir, InstallMask


class PackagePhase(CompositeTask):
    """
    Invokes the package phase and handles PKG_INSTALL_MASK.
    """

    __slots__ = (
        "actionmap",
        "fd_pipes",
        "logfile",
        "settings",
        "_pkg_install_mask",
        "_proot",
    )

    _shell_binary = portage.const.BASH_BINARY

    def _start(self):
        try:
            with io.open(
                _unicode_encode(
                    os.path.join(
                        self.settings["PORTAGE_BUILDDIR"],
                        "build-info",
                        "PKG_INSTALL_MASK",
                    ),
                    encoding=_encodings["fs"],
                    errors="strict",
                ),
                mode="r",
                encoding=_encodings["repo.content"],
                errors="replace",
            ) as f:
                self._pkg_install_mask = InstallMask(f.read())
        except EnvironmentError:
            self._pkg_install_mask = None
        if self._pkg_install_mask:
            self._proot = os.path.join(self.settings["T"], "packaging")
            self._start_task(
                SpawnProcess(
                    args=[
                        self._shell_binary,
                        "-e",
                        "-c",
                        (
                            "rm -rf {PROOT}; "
                            'cp -pPR $(cp --help | grep -q -- "^[[:space:]]*-l," && echo -l)'
                            ' "${{D}}" {PROOT}'
                        ).format(PROOT=portage._shell_quote(self._proot)),
                    ],
                    background=self.background,
                    env=self.settings.environ(),
                    scheduler=self.scheduler,
                    logfile=self.logfile,
                ),
                self._copy_proot_exit,
            )
        else:
            self._proot = self.settings["D"]
            self._start_package_phase()

    def _copy_proot_exit(self, proc):
        if self._default_exit(proc) != os.EX_OK:
            self.wait()
        else:
            self._start_task(
                AsyncFunction(
                    target=install_mask_dir,
                    args=(
                        os.path.join(
                            self._proot, self.settings["EPREFIX"].lstrip(os.sep)
                        ),
                        self._pkg_install_mask,
                    ),
                ),
                self._pkg_install_mask_exit,
            )

    def _pkg_install_mask_exit(self, proc):
        if self._default_exit(proc) != os.EX_OK:
            self.wait()
        else:
            self._start_package_phase()

    def _start_package_phase(self):
        ebuild_process = EbuildProcess(
            actionmap=self.actionmap,
            background=self.background,
            fd_pipes=self.fd_pipes,
            logfile=self.logfile,
            phase="package",
            scheduler=self.scheduler,
            settings=self.settings,
        )

        if self._pkg_install_mask:
            d_orig = self.settings["D"]
            try:
                self.settings["D"] = self._proot
                self._start_task(ebuild_process, self._pkg_install_mask_cleanup)
            finally:
                self.settings["D"] = d_orig
        else:
            self._start_task(ebuild_process, self._default_final_exit)

    def _pkg_install_mask_cleanup(self, proc):
        if self._default_exit(proc) != os.EX_OK:
            self.wait()
        else:
            self._start_task(
                SpawnProcess(
                    args=["rm", "-rf", self._proot],
                    background=self.background,
                    env=self.settings.environ(),
                    scheduler=self.scheduler,
                    logfile=self.logfile,
                ),
                self._default_final_exit,
            )



# Copyright 2010-2018 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import errno
import logging
import pickle
from portage import os
from portage.exception import TryAgain
from portage.localization import _
from portage.locks import lockfile, unlockfile
from portage.util import writemsg_level
from _emerge.FifoIpcDaemon import FifoIpcDaemon


class EbuildIpcDaemon(FifoIpcDaemon):
    """
    This class serves as an IPC daemon, which ebuild processes can use
    to communicate with portage's main python process.

    Here are a few possible uses:

    1) Robust subshell/subprocess die support. This allows the ebuild
       environment to reliably die without having to rely on signal IPC.

    2) Delegation of portageq calls to the main python process, eliminating
       performance and userpriv permission issues.

    3) Reliable ebuild termination in cases when the ebuild has accidentally
       left orphan processes running in the background (as in bug #278895).

    4) Detect cases in which bash has exited unexpectedly (as in bug #190128).
    """

    __slots__ = ("commands",)

    def _input_handler(self):
        # Read the whole pickle in a single atomic read() call.
        data = self._read_buf(self._files.pipe_in)
        if data is None:
            pass  # EAGAIN
        elif data:
            try:
                obj = pickle.loads(data)
            except SystemExit:
                raise
            except Exception:
                # The pickle module can raise practically
                # any exception when given corrupt data.
                pass
            else:

                self._reopen_input()

                cmd_key = obj[0]
                cmd_handler = self.commands[cmd_key]
                reply = cmd_handler(obj)
                try:
                    self._send_reply(reply)
                except OSError as e:
                    if e.errno == errno.ENXIO:
                        # This happens if the client side has been killed.
                        pass
                    else:
                        raise

                # Allow the command to execute hooks after its reply
                # has been sent. This hook is used by the 'exit'
                # command to kill the ebuild process. For some
                # reason, the ebuild-ipc helper hangs up the
                # ebuild process if it is waiting for a reply
                # when we try to kill the ebuild process.
                reply_hook = getattr(cmd_handler, "reply_hook", None)
                if reply_hook is not None:
                    reply_hook()

        else:  # EIO/POLLHUP
            # This can be triggered due to a race condition which happens when
            # the previous _reopen_input() call occurs before the writer has
            # closed the pipe (see bug #401919). It's not safe to re-open
            # without a lock here, since it's possible that another writer will
            # write something to the pipe just before we close it, and in that
            # case the write will be lost. Therefore, try for a non-blocking
            # lock, and only re-open the pipe if the lock is acquired.
            lock_filename = os.path.join(os.path.dirname(self.input_fifo), "lock")
            try:
                lock_obj = lockfile(lock_filename, unlinkfile=True, flags=os.O_NONBLOCK)
            except TryAgain:
                # We'll try again when another IO_HUP event arrives.
                pass
            else:
                try:
                    self._reopen_input()
                finally:
                    unlockfile(lock_obj)

    def _send_reply(self, reply):
        # File streams are in unbuffered mode since we do atomic
        # read and write of whole pickles. Use non-blocking mode so
        # we don't hang if the client is killed before we can send
        # the reply. We rely on the client opening the other side
        # of this fifo before it sends its request, since otherwise
        # we'd have a race condition with this open call raising
        # ENXIO if the client hasn't opened the fifo yet.
        try:
            output_fd = os.open(self.output_fifo, os.O_WRONLY | os.O_NONBLOCK)
            try:
                os.write(output_fd, pickle.dumps(reply))
            finally:
                os.close(output_fd)
        except OSError as e:
            # This probably means that the client has been killed,
            # which causes open to fail with ENXIO.
            writemsg_level(
                "!!! EbuildIpcDaemon %s: %s\n" % (_("failed to send reply"), e),
                level=logging.ERROR,
                noiselevel=-1,
            )



# Copyright 2005-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2
# Author(s): Nicholas Carpaski (carpaski@gentoo.org), Brian Harring (ferringb@gentoo.org)

__all__ = ["cache"]

import stat
import operator
import warnings
from portage.util import normalize_path
import errno
from portage.exception import FileNotFound, PermissionDenied
from portage import os
from portage import checksum
from portage import _shell_quote


class hashed_path:
    def __init__(self, location):
        self.location = location

    def __getattr__(self, attr):
        if attr == "mtime":
            # use stat.ST_MTIME; accessing .st_mtime gets you a float
            # depending on the python version, and int(float) introduces
            # some rounding issues that aren't present for people using
            # the straight c api.
            # thus use the defacto python compatibility work around;
            # access via index, which guarantees you get the raw int.
            try:
                self.mtime = obj = os.stat(self.location)[stat.ST_MTIME]
            except OSError as e:
                if e.errno in (errno.ENOENT, errno.ESTALE):
                    raise FileNotFound(self.location)
                elif e.errno == PermissionDenied.errno:
                    raise PermissionDenied(self.location)
                raise
            return obj
        if not attr.islower():
            # we don't care to allow .mD5 as an alias for .md5
            raise AttributeError(attr)
        hashname = attr.upper()
        if hashname not in checksum.get_valid_checksum_keys():
            raise AttributeError(attr)
        val = checksum.perform_checksum(self.location, hashname)[0]
        setattr(self, attr, val)
        return val

    def __repr__(self):
        return f"<portage.eclass_cache.hashed_path('{self.location}')>"


class cache:
    """
    Maintains the cache information about eclasses used in ebuild.
    """

    def __init__(self, porttree_root, overlays=None):
        if overlays is not None:
            warnings.warn(
                "overlays parameter of portage.eclass_cache.cache constructor is deprecated and no longer used",
                DeprecationWarning,
                stacklevel=2,
            )

        self.eclasses = {}  # {"Name": hashed_path}
        self._eclass_locations = {}
        self._eclass_locations_str = None

        # screw with the porttree ordering, w/out having bash inherit match it, and I'll hurt you.
        # ~harring
        if porttree_root:
            self.porttree_root = porttree_root
            self.porttrees = (normalize_path(self.porttree_root),)
            self._master_eclass_root = os.path.join(self.porttrees[0], "eclass")
            self.update_eclasses()
        else:
            self.porttree_root = None
            self.porttrees = ()
            self._master_eclass_root = None

    def copy(self):
        return self.__copy__()

    def __copy__(self):
        result = self.__class__(None)
        result.eclasses = self.eclasses.copy()
        result._eclass_locations = self._eclass_locations.copy()
        result.porttree_root = self.porttree_root
        result.porttrees = self.porttrees
        result._master_eclass_root = self._master_eclass_root
        return result

    def append(self, other):
        """
        Append another instance to this instance. This will cause eclasses
        from the other instance to override any eclasses from this instance
        that have the same name.
        """
        if not isinstance(other, self.__class__):
            raise TypeError(f"expected type {self.__class__}, got {type(other)}")
        self.porttrees = self.porttrees + other.porttrees
        self.eclasses.update(other.eclasses)
        self._eclass_locations.update(other._eclass_locations)
        self._eclass_locations_str = None

    def update_eclasses(self):
        self.eclasses = {}
        self._eclass_locations = {}
        master_eclasses = {}
        eclass_len = len(".eclass")
        ignored_listdir_errnos = (errno.ENOENT, errno.ENOTDIR)
        for x in (normalize_path(os.path.join(y, "eclass")) for y in self.porttrees):
            try:
                eclass_filenames = os.listdir(x)
            except OSError as e:
                if e.errno in ignored_listdir_errnos:
                    del e
                    continue
                elif e.errno == PermissionDenied.errno:
                    raise PermissionDenied(x)
                raise
            for y in eclass_filenames:
                if not y.endswith(".eclass"):
                    continue
                obj = hashed_path(os.path.join(x, y))
                obj.eclass_dir = x
                try:
                    mtime = obj.mtime
                except FileNotFound:
                    continue
                ys = y[:-eclass_len]
                if x == self._master_eclass_root:
                    master_eclasses[ys] = mtime
                    self.eclasses[ys] = obj
                    self._eclass_locations[ys] = x
                    continue

                master_mtime = master_eclasses.get(ys)
                if master_mtime is not None:
                    if master_mtime == mtime:
                        # It appears to be identical to the master,
                        # so prefer the master entry.
                        continue

                self.eclasses[ys] = obj
                self._eclass_locations[ys] = x

    def validate_and_rewrite_cache(self, ec_dict, chf_type, stores_paths):
        """
        This will return an empty dict if the ec_dict parameter happens
        to be empty, therefore callers must take care to distinguish
        between empty dict and None return values.
        """
        if not isinstance(ec_dict, dict):
            return None
        our_getter = operator.attrgetter(chf_type)
        cache_getter = lambda x: x
        if stores_paths:
            cache_getter = operator.itemgetter(1)
        d = {}
        for eclass, ec_data in ec_dict.items():
            cached_data = self.eclasses.get(eclass)
            if cached_data is None:
                return None
            if cache_getter(ec_data) != our_getter(cached_data):
                return None
            d[eclass] = cached_data
        return d

    def get_eclass_data(self, inherits):
        ec_dict = {x: self.eclasses[x] for x in inherits}
        return ec_dict

    @property
    def eclass_locations_string(self):
        if self._eclass_locations_str is None:
            self._eclass_locations_str = " ".join(
                _shell_quote(x) for x in reversed(self.porttrees)
            )
        return self._eclass_locations_str



# Copyright 2016 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import subprocess
import sys

import portage
from portage import os
from portage import _unicode_decode
from portage.const import PORTAGE_PYM_PATH, USER_CONFIG_PATH
from portage.process import find_binary
from portage.tests import TestCase
from portage.tests.resolver.ResolverPlayground import ResolverPlayground
from portage.util import ensure_dirs


class BlockerFileCollisionEmergeTestCase(TestCase):
    def testBlockerFileCollision(self):

        debug = False

        install_something = """
S="${WORKDIR}"

src_install() {
	einfo "installing something..."
	insinto /usr/lib
	echo "${PN}" > "${T}/file-collision"
	doins "${T}/file-collision"
}
"""

        ebuilds = {
            "dev-libs/A-1": {
                "EAPI": "6",
                "MISC_CONTENT": install_something,
                "RDEPEND": "!dev-libs/B",
            },
            "dev-libs/B-1": {
                "EAPI": "6",
                "MISC_CONTENT": install_something,
                "RDEPEND": "!dev-libs/A",
            },
        }

        playground = ResolverPlayground(ebuilds=ebuilds, debug=debug)
        settings = playground.settings
        eprefix = settings["EPREFIX"]
        eroot = settings["EROOT"]
        var_cache_edb = os.path.join(eprefix, "var", "cache", "edb")
        user_config_dir = os.path.join(eprefix, USER_CONFIG_PATH)

        portage_python = portage._python_interpreter
        emerge_cmd = (portage_python, "-b", "-Wd", os.path.join(self.bindir, "emerge"))

        file_collision = os.path.join(eroot, "usr/lib/file-collision")

        test_commands = (
            emerge_cmd
            + (
                "--oneshot",
                "dev-libs/A",
            ),
            (lambda: portage.util.grablines(file_collision) == ["A\n"],),
            emerge_cmd
            + (
                "--oneshot",
                "dev-libs/B",
            ),
            (lambda: portage.util.grablines(file_collision) == ["B\n"],),
            emerge_cmd
            + (
                "--oneshot",
                "dev-libs/A",
            ),
            (lambda: portage.util.grablines(file_collision) == ["A\n"],),
            ({"FEATURES": "parallel-install"},)
            + emerge_cmd
            + (
                "--oneshot",
                "dev-libs/B",
            ),
            (lambda: portage.util.grablines(file_collision) == ["B\n"],),
            ({"FEATURES": "parallel-install"},)
            + emerge_cmd
            + (
                "-Cq",
                "dev-libs/B",
            ),
            (lambda: not os.path.exists(file_collision),),
        )

        fake_bin = os.path.join(eprefix, "bin")
        portage_tmpdir = os.path.join(eprefix, "var", "tmp", "portage")
        profile_path = settings.profile_path

        path = os.environ.get("PATH")
        if path is not None and not path.strip():
            path = None
        if path is None:
            path = ""
        else:
            path = ":" + path
        path = fake_bin + path

        pythonpath = os.environ.get("PYTHONPATH")
        if pythonpath is not None and not pythonpath.strip():
            pythonpath = None
        if pythonpath is not None and pythonpath.split(":")[0] == PORTAGE_PYM_PATH:
            pass
        else:
            if pythonpath is None:
                pythonpath = ""
            else:
                pythonpath = ":" + pythonpath
            pythonpath = PORTAGE_PYM_PATH + pythonpath

        env = {
            "PORTAGE_OVERRIDE_EPREFIX": eprefix,
            "PATH": path,
            "PORTAGE_PYTHON": portage_python,
            "PORTAGE_REPOSITORIES": settings.repositories.config_string(),
            "PYTHONDONTWRITEBYTECODE": os.environ.get("PYTHONDONTWRITEBYTECODE", ""),
            "PYTHONPATH": pythonpath,
        }

        if "__PORTAGE_TEST_HARDLINK_LOCKS" in os.environ:
            env["__PORTAGE_TEST_HARDLINK_LOCKS"] = os.environ[
                "__PORTAGE_TEST_HARDLINK_LOCKS"
            ]

        dirs = [
            playground.distdir,
            fake_bin,
            portage_tmpdir,
            user_config_dir,
            var_cache_edb,
        ]
        true_symlinks = ["chown", "chgrp"]
        true_binary = find_binary("true")
        self.assertEqual(true_binary is None, False, "true command not found")
        try:
            for d in dirs:
                ensure_dirs(d)
            for x in true_symlinks:
                os.symlink(true_binary, os.path.join(fake_bin, x))
            with open(os.path.join(var_cache_edb, "counter"), "wb") as f:
                f.write(b"100")
            # non-empty system set keeps --depclean quiet
            with open(os.path.join(profile_path, "packages"), "w") as f:
                f.write("*dev-libs/token-system-pkg")

            if debug:
                # The subprocess inherits both stdout and stderr, for
                # debugging purposes.
                stdout = None
            else:
                # The subprocess inherits stderr so that any warnings
                # triggered by python -Wd will be visible.
                stdout = subprocess.PIPE

            for i, args in enumerate(test_commands):

                if hasattr(args[0], "__call__"):
                    self.assertTrue(args[0](), "callable at index %s failed" % (i,))
                    continue

                if isinstance(args[0], dict):
                    local_env = env.copy()
                    local_env.update(args[0])
                    args = args[1:]
                else:
                    local_env = env

                proc = subprocess.Popen(args, env=local_env, stdout=stdout)

                if debug:
                    proc.wait()
                else:
                    output = proc.stdout.readlines()
                    proc.wait()
                    proc.stdout.close()
                    if proc.returncode != os.EX_OK:
                        for line in output:
                            sys.stderr.write(_unicode_decode(line))

                self.assertEqual(
                    os.EX_OK, proc.returncode, "emerge failed with args %s" % (args,)
                )
        finally:
            playground.debug = False
            playground.cleanup()



# Copyright 2005-2020 Gentoo Authors
# Distributed under the terms of the GNU General Public License v2
# Author(s): Brian Harring (ferringb@gentoo.org)

from portage.cache import cache_errors
from portage.cache.cache_errors import InvalidRestriction
from portage.cache.mappings import ProtectedDict
import warnings
import operator


class database:
    # this is for metadata/cache transfer.
    # basically flags the cache needs be updated when transfered cache to cache.
    # leave this.

    complete_eclass_entries = True
    autocommits = False
    cleanse_keys = False
    serialize_eclasses = True
    validation_chf = "mtime"
    store_eclass_paths = True

    def __init__(self, location, label, auxdbkeys, readonly=False):
        """initialize the derived class; specifically, store label/keys"""
        self._known_keys = auxdbkeys
        self.location = location
        self.label = label
        self.readonly = readonly
        self.sync_rate = 0
        self.updates = 0

    def __getitem__(self, cpv):
        """set a cpv to values
        This shouldn't be overriden in derived classes since it handles the __eclasses__ conversion.
        that said, if the class handles it, they can override it."""
        if self.updates > self.sync_rate:
            self.commit()
            self.updates = 0
        d = self._getitem(cpv)

        try:
            chf_types = self.chf_types
        except AttributeError:
            chf_types = (self.validation_chf,)

        if self.serialize_eclasses and "_eclasses_" in d:
            for chf_type in chf_types:
                if "_%s_" % chf_type not in d:
                    # Skip the reconstruct_eclasses call, since it's
                    # a waste of time if it contains a different chf_type
                    # than the current one. In the past, it was possible
                    # for reconstruct_eclasses called with chf_type='md5'
                    # to "successfully" return invalid data here, because
                    # it was unable to distinguish between md5 data and
                    # mtime data.
                    continue
                try:
                    d["_eclasses_"] = reconstruct_eclasses(
                        cpv, d["_eclasses_"], chf_type, paths=self.store_eclass_paths
                    )
                except cache_errors.CacheCorruption:
                    if chf_type is chf_types[-1]:
                        raise
                else:
                    break
            else:
                raise cache_errors.CacheCorruption(
                    cpv, "entry does not contain a recognized chf_type"
                )

        elif "_eclasses_" not in d:
            d["_eclasses_"] = {}
        # Never return INHERITED, since portdbapi.aux_get() will
        # generate it automatically from _eclasses_, and we want
        # to omit it in comparisons between cache entries like
        # those that egencache uses to avoid redundant writes.
        d.pop("INHERITED", None)

        mtime_required = not any(d.get("_%s_" % x) for x in chf_types if x != "mtime")

        mtime = d.get("_mtime_")
        if not mtime:
            if mtime_required:
                raise cache_errors.CacheCorruption(cpv, "_mtime_ field is missing")
            d.pop("_mtime_", None)
        else:
            try:
                mtime = int(mtime)
            except ValueError:
                raise cache_errors.CacheCorruption(
                    cpv, "_mtime_ conversion to int failed: %s" % (mtime,)
                )
            d["_mtime_"] = mtime
        return d

    def _getitem(self, cpv):
        """get cpv's values.
        override this in derived classess"""
        raise NotImplementedError

    @staticmethod
    def _internal_eclasses(extern_ec_dict, chf_type, paths):
        """
        When serialize_eclasses is False, we have to convert an external
        eclass dict containing hashed_path objects into an appropriate
        internal dict containing values of chf_type (and eclass dirs
        if store_eclass_paths is True).
        """
        if not extern_ec_dict:
            return extern_ec_dict
        chf_getter = operator.attrgetter(chf_type)
        if paths:
            intern_ec_dict = dict(
                (k, (v.eclass_dir, chf_getter(v))) for k, v in extern_ec_dict.items()
            )
        else:
            intern_ec_dict = dict((k, chf_getter(v)) for k, v in extern_ec_dict.items())
        return intern_ec_dict

    def __setitem__(self, cpv, values):
        """set a cpv to values
        This shouldn't be overriden in derived classes since it handles the readonly checks"""
        if self.readonly:
            raise cache_errors.ReadOnlyRestriction()
        d = None
        if self.cleanse_keys:
            d = ProtectedDict(values)
            for k, v in list(item for item in d.items() if item[0] != "_eclasses_"):
                if not v:
                    del d[k]
        if "_eclasses_" in values:
            if d is None:
                d = ProtectedDict(values)
            if self.serialize_eclasses:
                d["_eclasses_"] = serialize_eclasses(
                    d["_eclasses_"], self.validation_chf, paths=self.store_eclass_paths
                )
            else:
                d["_eclasses_"] = self._internal_eclasses(
                    d["_eclasses_"], self.validation_chf, self.store_eclass_paths
                )
        elif d is None:
            d = values
        self._setitem(cpv, d)
        if not self.autocommits:
            self.updates += 1
            if self.updates > self.sync_rate:
                self.commit()
                self.updates = 0

    def _setitem(self, name, values):
        """__setitem__ calls this after readonly checks.  override it in derived classes
        note _eclassees_ key *must* be handled"""
        raise NotImplementedError

    def __delitem__(self, cpv):
        """delete a key from the cache.
        This shouldn't be overriden in derived classes since it handles the readonly checks"""
        if self.readonly:
            raise cache_errors.ReadOnlyRestriction()
        if not self.autocommits:
            self.updates += 1
        self._delitem(cpv)
        if self.updates > self.sync_rate:
            self.commit()
            self.updates = 0

    def _delitem(self, cpv):
        """__delitem__ calls this after readonly checks.  override it in derived classes"""
        raise NotImplementedError

    def has_key(self, cpv):
        return cpv in self

    def iterkeys(self):
        return iter(self)

    def iteritems(self):
        for x in self:
            yield (x, self[x])

    def sync(self, rate=0):
        self.sync_rate = rate
        if rate == 0:
            self.commit()

    def commit(self):
        if not self.autocommits:
            raise NotImplementedError(self)

    def __del__(self):
        # This used to be handled by an atexit hook that called
        # close_portdbapi_caches() for all portdbapi instances, but that was
        # prone to memory leaks for API consumers that needed to create/destroy
        # many portdbapi instances. So, instead we rely on __del__.
        self.sync()

    def __contains__(self, cpv):
        """This method should always be overridden.  It is provided only for
        backward compatibility with modules that override has_key instead.  It
        will automatically raise a NotImplementedError if has_key has not been
        overridden."""
        if self.has_key is database.has_key:
            # prevent a possible recursive loop
            raise NotImplementedError
        warnings.warn(
            "portage.cache.template.database.has_key() is "
            "deprecated, override __contains__ instead",
            DeprecationWarning,
        )
        return self.has_key(cpv)

    def __iter__(self):
        """This method should always be overridden.  It is provided only for
        backward compatibility with modules that override iterkeys instead.  It
        will automatically raise a NotImplementedError if iterkeys has not been
        overridden."""
        if self.iterkeys is database.iterkeys:
            # prevent a possible recursive loop
            raise NotImplementedError(self)
        return iter(self.keys())

    def get(self, k, x=None):
        try:
            return self[k]
        except KeyError:
            return x

    def validate_entry(self, entry, ebuild_hash, eclass_db):
        try:
            chf_types = self.chf_types
        except AttributeError:
            chf_types = (self.validation_chf,)

        for chf_type in chf_types:
            if self._validate_entry(chf_type, entry, ebuild_hash, eclass_db):
                return True

        return False

    def _validate_entry(self, chf_type, entry, ebuild_hash, eclass_db):
        hash_key = "_%s_" % chf_type
        try:
            entry_hash = entry[hash_key]
        except KeyError:
            return False
        else:
            if entry_hash != getattr(ebuild_hash, chf_type):
                return False
        update = eclass_db.validate_and_rewrite_cache(
            entry["_eclasses_"], chf_type, self.store_eclass_paths
        )
        if update is None:
            return False
        if update:
            entry["_eclasses_"] = update
        return True

    def get_matches(self, match_dict):
        """generic function for walking the entire cache db, matching restrictions to
        filter what cpv's are returned.  Derived classes should override this if they
        can implement a faster method then pulling each cpv:values, and checking it.

        For example, RDBMS derived classes should push the matching logic down to the
        actual RDBM."""

        import re

        restricts = {}
        for key, match in match_dict.items():
            # XXX this sucks.
            try:
                if isinstance(match, str):
                    restricts[key] = re.compile(match).match
                else:
                    restricts[key] = re.compile(match[0], match[1]).match
            except re.error as e:
                raise InvalidRestriction(key, match, e)
            if key not in self.__known_keys:
                raise InvalidRestriction(key, match, "Key isn't valid")

        for cpv in self:
            cont = True
            vals = self[cpv]
            for key, match in restricts.items():
                if not match(vals[key]):
                    cont = False
                    break
            if cont:
                yield cpv

    keys = __iter__
    items = iteritems


_keysorter = operator.itemgetter(0)


def serialize_eclasses(eclass_dict, chf_type="mtime", paths=True):
    """takes a dict, returns a string representing said dict"""
    """The "new format", which causes older versions of <portage-2.1.2 to
	traceback with a ValueError due to failed int() conversion.  This format
	isn't currently written, but the capability to read it is already built
	in.
	return "\t".join(["%s\t%s" % (k, str(v)) \
		for k, v in eclass_dict.iteritems()])
	"""
    if not eclass_dict:
        return ""
    getter = operator.attrgetter(chf_type)
    if paths:
        return "\t".join(
            "%s\t%s\t%s" % (k, v.eclass_dir, getter(v))
            for k, v in sorted(eclass_dict.items(), key=_keysorter)
        )
    return "\t".join(
        "%s\t%s" % (k, getter(v))
        for k, v in sorted(eclass_dict.items(), key=_keysorter)
    )


def _md5_deserializer(md5):
    """
    Without this validation, it's possible for reconstruct_eclasses to
    mistakenly interpret mtime data as md5 data, and return an invalid
    data structure containing strings where ints are expected.
    """
    if len(md5) != 32:
        raise ValueError("expected 32 hex digits")
    return md5


_chf_deserializers = {
    "md5": _md5_deserializer,
    "mtime": int,
}


def reconstruct_eclasses(cpv, eclass_string, chf_type="mtime", paths=True):
    """returns a dict when handed a string generated by serialize_eclasses"""
    eclasses = eclass_string.rstrip().lstrip().split("\t")
    if eclasses == [""]:
        # occasionally this occurs in the fs backends.  they suck.
        return {}

    converter = _chf_deserializers.get(chf_type, lambda x: x)

    if paths:
        if len(eclasses) % 3 != 0:
            raise cache_errors.CacheCorruption(
                cpv, "_eclasses_ was of invalid len %i" % len(eclasses)
            )
    elif len(eclasses) % 2 != 0:
        raise cache_errors.CacheCorruption(
            cpv, "_eclasses_ was of invalid len %i" % len(eclasses)
        )
    d = {}
    try:
        i = iter(eclasses)
        if paths:
            # The old format contains paths that will be discarded.
            for name, path, val in zip(i, i, i):
                d[name] = (path, converter(val))
        else:
            for name, val in zip(i, i):
                d[name] = converter(val)
    except IndexError:
        raise cache_errors.CacheCorruption(
            cpv, "_eclasses_ was of invalid len %i" % len(eclasses)
        )
    except ValueError:
        raise cache_errors.CacheCorruption(
            cpv, "_eclasses_ not valid for chf_type {}".format(chf_type)
        )
    del eclasses
    return d



# Copyright 1999-2012 Gentoo Foundation
# Distributed under the terms of the GNU General Public License v2

import errno
import logging
import re
import stat
import subprocess

import portage
from portage import os


def chk_updated_info_files(root, infodirs, prev_mtimes):

    if os.path.exists("/usr/bin/install-info"):
        out = portage.output.EOutput()
        regen_infodirs = []
        for z in infodirs:
            if z == "":
                continue
            inforoot = portage.util.normalize_path(root + z)
            if os.path.isdir(inforoot) and not [
                x for x in os.listdir(inforoot) if x.startswith(".keepinfodir")
            ]:
                infomtime = os.stat(inforoot)[stat.ST_MTIME]
                if inforoot not in prev_mtimes or prev_mtimes[inforoot] != infomtime:
                    regen_infodirs.append(inforoot)

        if not regen_infodirs:
            portage.util.writemsg_stdout("\n")
            if portage.util.noiselimit >= 0:
                out.einfo("GNU info directory index is up-to-date.")
        else:
            portage.util.writemsg_stdout("\n")
            if portage.util.noiselimit >= 0:
                out.einfo("Regenerating GNU info directory index...")

            dir_extensions = ("", ".gz", ".bz2")
            icount = 0
            badcount = 0
            errmsg = ""
            for inforoot in regen_infodirs:
                if inforoot == "":
                    continue

                if not os.path.isdir(inforoot) or not os.access(inforoot, os.W_OK):
                    continue

                file_list = os.listdir(inforoot)
                file_list.sort()
                dir_file = os.path.join(inforoot, "dir")
                moved_old_dir = False
                processed_count = 0
                for x in file_list:
                    if x.startswith(".") or os.path.isdir(os.path.join(inforoot, x)):
                        continue
                    if x.startswith("dir"):
                        skip = False
                        for ext in dir_extensions:
                            if x == "dir" + ext or x == "dir" + ext + ".old":
                                skip = True
                                break
                        if skip:
                            continue
                    if processed_count == 0:
                        for ext in dir_extensions:
                            try:
                                os.rename(dir_file + ext, dir_file + ext + ".old")
                                moved_old_dir = True
                            except EnvironmentError as e:
                                if e.errno != errno.ENOENT:
                                    raise
                                del e
                    processed_count += 1
                    try:
                        proc = subprocess.Popen(
                            [
                                "/usr/bin/install-info",
                                "--dir-file=%s" % os.path.join(inforoot, "dir"),
                                os.path.join(inforoot, x),
                            ],
                            env=dict(os.environ, LANG="C", LANGUAGE="C"),
                            stdout=subprocess.PIPE,
                            stderr=subprocess.STDOUT,
                        )
                    except OSError:
                        myso = None
                    else:
                        myso = portage._unicode_decode(proc.communicate()[0]).rstrip(
                            "\n"
                        )
                        proc.wait()
                    existsstr = "already exists, for file `"
                    if myso:
                        if re.search(existsstr, myso):
                            # Already exists... Don't increment the count for this.
                            pass
                        elif (
                            myso[:44] == "install-info: warning: no info dir entry in "
                        ):
                            # This info file doesn't contain a DIR-header: install-info produces this
                            # (harmless) warning (the --quiet switch doesn't seem to work).
                            # Don't increment the count for this.
                            pass
                        else:
                            badcount += 1
                            errmsg += myso + "\n"
                    icount += 1

                if moved_old_dir and not os.path.exists(dir_file):
                    # We didn't generate a new dir file, so put the old file
                    # back where it was originally found.
                    for ext in dir_extensions:
                        try:
                            os.rename(dir_file + ext + ".old", dir_file + ext)
                        except EnvironmentError as e:
                            if e.errno != errno.ENOENT:
                                raise
                            del e

                # Clean dir.old cruft so that they don't prevent
                # unmerge of otherwise empty directories.
                for ext in dir_extensions:
                    try:
                        os.unlink(dir_file + ext + ".old")
                    except EnvironmentError as e:
                        if e.errno != errno.ENOENT:
                            raise
                        del e

                # update mtime so we can potentially avoid regenerating.
                prev_mtimes[inforoot] = os.stat(inforoot)[stat.ST_MTIME]

            if badcount:
                out.eerror("Processed %d info files; %d errors." % (icount, badcount))
                portage.util.writemsg_level(errmsg, level=logging.ERROR, noiselevel=-1)
            else:
                if icount > 0 and portage.util.noiselimit >= 0:
                    out.einfo("Processed %d info files." % (icount,))
